/*
 * Conditional expression grammar and parser.
 *
 * David Leonard, 2001. Public domain.
 */

%{
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <err.h>

#include "hash.h"
#include "scope.h"
#include "value.h"
#include "func.h"

#define ARGS value_t *v[] 
#define APPLY(op, result)	\
	do {								\
		int i;							\
		apply(op, result, v, sizeof v/sizeof v[0]);		\
		for (i = 0; i < sizeof v/sizeof v[0]; i++)		\
			value_free(*v[i]);				\
	} while (0)

#define BAD(v, m)	(v).v_type = UNK; yyerror(m)

static value_t intzero = { INT };
void applyvals(char *, value_t *, values_t *);

int parse_errors;
value_t parse_result;
scope_t *parse_scope;

int yylex(void);
void yyerror(const char *fmt, ...);

%}

%union{
	value_t value;
	values_t *values;
	char	ident[32];
}

%token	'+' '-' '/' '*' '(' ')' '<' '>' '=' ','
%token	GE LE NE AND OR NOT BEGINS MATCHES
%token	VALUE IDENT THEN DEFINED

%type	<value>	expr VALUE
%type	<values> values
%type	<ident>	IDENT

%left	','
%left	OR
%left	AND
%left   BEGINS MATCHES
%left	'=' NE
%left	'<' LE '>' GE
%left	'+' '-'
%left	'*' '/' 
%right	NOT
%left	'('

%%

top	:	expr THEN	{ parse_result = $1; }
	;

expr	:	'(' expr ')'	{ $$ = $2; }
	|	VALUE		{ $$ = $1; }
	|	IDENT '(' values ')'
				{ applyvals($1, &$$, $3); }
	|	IDENT 		{ apply($1, &$$, NULL, 0); }
	|	IDENT '(' ')'	{ apply($1, &$$, NULL, 0); }	/* illegal? */
	|	DEFINED '(' IDENT ')'
				{ $$.v_type = LOG; 
				  $$.v_log = (scope_lookup(parse_scope,
				      $3) != NULL); }
	|	expr '+' expr	{ ARGS = { &$1, &$3 }; APPLY("+", &$$); }
	|	expr '-' expr	{ ARGS = { &$1, &$3 }; APPLY("-", &$$); }
	|	expr '*' expr	{ ARGS = { &$1, &$3 }; APPLY("*", &$$); }
	|	expr '/' expr	{ ARGS = { &$1, &$3 }; APPLY("/", &$$); }
	|	expr '=' expr	{ ARGS = { &$1, &$3 }; APPLY("=", &$$); }
	|	expr '<' expr	{ ARGS = { &$1, &$3 }; APPLY("<", &$$); }
	|	expr '>' expr	{ ARGS = { &$1, &$3 }; APPLY(">", &$$); }
	|	expr GE  expr	{ ARGS = { &$1, &$3 }; APPLY(">=", &$$); }
	|	expr LE  expr	{ ARGS = { &$1, &$3 }; APPLY("<=", &$$); }
	|	expr NE  expr	{ ARGS = { &$1, &$3 }; APPLY("<>", &$$); }
	|	expr AND  expr	{ ARGS = { &$1, &$3 }; APPLY("&&", &$$); }
	|	expr OR  expr	{ ARGS = { &$1, &$3 }; APPLY("||", &$$); }
	|	expr BEGINS expr { ARGS ={ &$1, &$3 }; APPLY("+begins", &$$); }
	|	expr MATCHES expr { ARGS={ &$1, &$3 }; APPLY("+matches", &$$); }
	|	NOT  expr	{ ARGS = { &$2 }; APPLY("!", &$$); }
	|	'-' expr	{ ARGS = { &intzero, &$2}; APPLY("-", &$$); }
			%prec NOT
	|	'+' expr	{ ARGS = { &intzero, &$2}; APPLY("+", &$$); }
			%prec NOT
	|	'(' expr error	{ BAD($$, "unclosed parenthesis"); }
	;

values	:	expr		{ $$ = (values_t *)malloc(sizeof (values_t));
				  $$->next = NULL;
				  $$->val = $1; }
	|	values ',' expr	{ $$ = (values_t *)malloc(sizeof (values_t));
				  $$->next = $1;
				  $$->val = $3; }
	;

%%

void
applyvals(name, dest, vals)
	char *name;
	value_t *dest;
	values_t *vals;
{
	int i = 0, nvals;
	values_t *v;
	value_t **va;

	/* Count the number of values */
	for (nvals = 0, v = vals; v; v = v->next, nvals++)
		;
	if (nvals)
		va = (value_t **)malloc(nvals * sizeof (value_t *));
	else
		va = NULL;
	for (i = nvals, v = vals; v; v = v->next, i--)
		va[i - 1] = &v->val;
	apply(name, dest, va, nvals);
	if (va != NULL)
		free(va);
	while (vals) {
		v = vals;
		vals = vals->next;
		value_free(v->val);
		free(v);
	}
}

void
yyerror(const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	if (parse_scope)
		fprintf(stderr, "%s: %d: ", parse_scope->filename, 
			parse_scope->lineno);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");
	parse_errors++;
}
