/*
 * Progresh preprocessor
 *
 * David Leonard, 2001. Public domain.
 */

#include <stdio.h>
#include <err.h>
#include <stdlib.h>
#include <string.h>

#include "ifelse.h"
#include "buf.h"
#include "hash.h"
#include "scope.h"
#include "ifexpr.h"
#include "tokenize.h"
#include "keyword.h"

#define isspace(c) ((c) == ' ' || (c) == '\t' || (c) == '\n' || (c) == '\r')

static char *OPSYS	= "UNIX";
static char *WINDOWSYS	= "TTY";


scope_t *globalscope;

int oneline(scope_t *, struct buf *);
int expand(scope_t *, struct buf *, int, struct buf *, int, int);
void loadfile(scope_t *, char **, struct buf *, int);
int directive(scope_t *, struct buf *, int);

int batchmode = 0;		/* -b flag */
int sequenceno = 0;		/* for {&sequence} */

/* Read one line from the scope file, consuming ~\n and eliminating comments */
int
oneline(scope, buf)
	scope_t *scope;
	struct buf *buf;
{
	FILE *f = scope->file;
	int c, lastc = '\0';
	int commentdepth = 0;
	int startlen = buf->len;

	while (!feof(f)) {
		c = fgetc(f);
		if (c == EOF) {
			if (commentdepth)
				warnx("%s: unclosed comment at eof",
					scope->filename);
			break;
		}

		if (lastc == '/') {
			if (c == '*')
				commentdepth++;
			else
				if (commentdepth == 0)
					buf_append(buf, '/');
		} 
		if (commentdepth > 0 && lastc == '*' && c == '/') {
			commentdepth--;
			continue;
		}
		lastc = c;
		if (c == '/')
			continue;

		if (commentdepth > 0)
			continue;

		if (c == '~' || c == '\\') {
			int d = fgetc(f);
			if (d == '\n') {
				scope->lineno++;
				continue;
			} 
			buf_append(buf, c);
			buf_append(buf, d);
			continue;
		}
		buf_append(buf, c);
		if (c == '\n') {
			scope->lineno++;
			break;
		}
	}
	return buf->len - startlen;
}

/*
 * Expands part of ibuf up to the next closing '}', output into obuf.
 * Assumes character at start is a '{'
 * Returns the position in ibuf after the closing '}'.
 */
int
expand(scope, ibuf, start, obuf, linesync, all)
	scope_t *scope;
	struct buf *ibuf, *obuf;
	int start;
	int linesync;
	int all;
{
	int i;
	struct buf inside, val;
	char **args;
	char buf[256];

	for (i = start; i < ibuf->len; i++) {
		char c = ibuf->data[i];
								/* { */
		if (c == '}' && !all)
			break;
		if (c == '~') {
			/* Copy next input char literally */
			i++;
			buf_append(obuf, ibuf->data[i]);
			continue;
		}
		if (c != '{') {					/* } */
			buf_append(obuf, ibuf->data[i]);
			continue;
		}
		buf_init(&inside);
		/* Recursive expand */
		i = expand(scope, ibuf, i + 1, &inside, 0, 0);
		args = tokenize(scope, &inside);
#ifdef DEBUG
{int j;
fprintf(stderr, "expanding {");
for (j = 0; args[j]; j++)
 fprintf(stderr, "%s%s", j==0?"":" ",args[j]);
fprintf(stderr, "}\n");
}
#endif

		val.data = NULL;
		val.alloc = 0;
		if (args[0] == NULL) {
			/* ignore {} */
#ifdef DEBUG
			warnx("%s:%d: empty reference text",
				scope->filename, scope->lineno);
#endif
		} else if (args[0][0] >= '1' && args[0][0] <= '9') {
			int n = atoi(args[0]);
			if (n >= 1 && n < scope->nargs)
				val.data = scope->args[n];
#ifdef DEBUG
			else
				warnx("%s:%d: invalid argument number {%s}",
				       scope->filename, scope->lineno, args[0]);
#endif
		} else if (strcmp(args[0], "*") == 0) {
			int n;
			buf_init(&val);
			for (n = 1; n < scope->nargs; n++) {
				if (n > 1)
					buf_append(&val, ' ');
				buf_addstr(&val, scope->args[n]);
			}
		} else if (args[0][0] == '&') {
			char *w = args[0];
			val.data = scope_lookup(scope, w + 1);
			if (val.data != NULL) {
				/* overridden variable */
			} else if (keyword_match(w, "&file-name") == 0) {
				val.data = scope->filename;
			} else if (keyword_match(w, "&line-numbe|r") == 0) {
				sprintf(buf, "%d", scope->lineno);
				val.data = buf;
			} else if (keyword_match(w, "&batch|-mode") == 0) {
				sprintf(buf, "%s", batchmode ? "YES" : "NO");
				val.data = buf;
			} else if (keyword_match(w, "&window-sys|tem") == 0) {
				val.data = WINDOWSYS;
			} else if (keyword_match(w, "&opsys") == 0) {
				val.data = OPSYS;
			} else if (keyword_match(w, "&sequence") == 0) {
				sprintf(buf, "%d", sequenceno++);
				val.data = buf;
			}
#ifdef DEBUG
			else
				warnx("%s:%d: undefined {%s}",
				    scope->filename, scope->lineno, w);
#endif
		} else 
			loadfile(scope, args, obuf, linesync);
		if (val.data != NULL) {
			val.len = strlen(val.data);
			expand(scope, &val, 0, obuf, 0, 1);
		}
		if (val.alloc)
			buf_free(&val);
		free(args);
		buf_free(&inside);
	}
	return i;
}

/* Emit a #line directive to synchronise second stage compiler. */
void
addsync(scope, obuf)
	scope_t *scope;
	struct buf *obuf;
{
	char sync[80];
	sprintf(sync, "#line %d %s\n",
		scope->lineno, scope->filename);
	buf_addstr(obuf, sync);
}

/* Expanding a file include reference into obuf */
void
loadfile(parentscope, args, obuf, linesync)
	scope_t *parentscope;
	char **args;
	struct buf *obuf;
	int linesync;
{
	struct buf line, dctv;
	scope_t *scope;
	int i;
	int ampstarts = -1;
	int startline;

	/* Enter a new scope */
	scope = new_scope();
	scope->ifdepth = 0;

	/* Look for named arguments, define as locals */
	if (args[0] == NULL)
		i = 0;
	else
		for (i = 1; args[i] != NULL; i++) {
			if (args[i][0] == '&') {
				char *eq = strchr(args[i], '=');
				*eq++ = '\0';
				scope_define(scope, args[i]+1, eq);
				args[i] = eq;
			}
		}
	scope->args = args;
	scope->nargs = i;

	if (args[0] == NULL || strcmp(args[0], "-") == 0) {
		scope->filename = "-";
		scope->file = stdin;
	} else {
		scope->filename = args[0];
		scope->file = fopen(scope->filename, "r");
		if (scope->file == NULL) {
			if (parentscope == NULL)
				warn("%s", scope->filename);
			else
				warn("%s:%d: %s",
					parentscope->filename,
					parentscope->lineno,
					scope->filename);
			free_scope(scope);
			return;
		}
	}
	scope->lineno = 0;

	scope->outer = parentscope;
	if (scope->outer == NULL) 
		globalscope = scope;

	buf_init(&line);
	buf_init(&dctv);
	while (!feof(scope->file)) {
		int i;
		if (oneline(scope, &line) != 0) {
			int braces = 0;
			/* Make sure braces match up */
			for (i = 0; i < line.len; i++)
				if (line.data[i] == '~')
					i++;
				else if (line.data[i] == '{')
					braces++;
				else if (line.data[i] == '}' && braces > 0)
					braces--;
			if (braces > 0)
				/* Read more lines if braces didn't match up */
				continue;
		} 

		/* Check for end of file */
		if (line.len == 0)
			break;

		if (dctv.len == 0) {
			/* Look for an ampersand starting the line */
			for (i = 0; i < line.len; i++) 
				if (!isspace(line.data[i]))
					break;
			ampstarts = (i < line.len && line.data[i] == '&') 
					? i : -1;
			startline = scope->lineno;
		}

#ifdef DEBUG
fprintf(stderr, "line = '%.*s'\n", line.len, line.data);
#endif
		/* Expand the braces in the line - append to dctv */
		expand(scope, &line, 0, &dctv, linesync && ampstarts != -1, 1);
		line.len = 0;

#ifdef DEBUG
fprintf(stderr, "dctv = '%.*s'\n", dctv.len, dctv.data);
fprintf(stderr, "ampstarts = %d\n", ampstarts);
#endif

		/* Check if we are handling a directive */
		if (ampstarts != -1) {
			if (directive(scope, &dctv, ampstarts))
				/* Needs more lines added to dctv */
				continue;
		}
		else if (if_active()) {
			if (linesync) 
				addsync(scope, obuf);
			buf_cat(obuf, &dctv);
		}
		dctv.len = 0;
	}
	buf_free(&dctv);
	buf_free(&line);

	if (linesync) 
		addsync(scope, obuf);

	if (scope->ifdepth > 0)
		warnx("%s:%d: unclosed &if",
			scope->filename, scope->lineno);
	while (scope->ifdepth > 0) {
		if_endif();
		scope->ifdepth--;
	}
}

/*
 * Process a '&' command/directive 
 * Returns 1 if directive needs more data to continue.
 */
int
directive(scope, line, start)
	scope_t *scope;
	struct buf *line;
	int start;
{
	int i;
	char save;
	enum { GLOB, SCOP, UNDEF, IF, ELSEIF, ELSE, ENDIF, MESSAGE } d;
	char *op;

	line->data[line->len] = '\0';
	for (i = start; i < line->len; i++)
		if (isspace(line->data[i]))
			break;
	save = line->data[i];
	line->data[i] = '\0';
	op = line->data + start;
	if (keyword_match(op, "&glob|al-define") == 0)		d = GLOB;
	else if (keyword_match(op, "&scop|ed-define") == 0)	d = SCOP;
	else if (keyword_match(op, "&undef|ine") == 0)		d = UNDEF;
	else if (keyword_match(op, "&if") == 0)			d = IF;
	else if (keyword_match(op, "&elseif") == 0)		d = ELSEIF;
	else if (keyword_match(op, "&else") == 0)		d = ELSE;
	else if (keyword_match(op, "&endif") == 0)		d = ENDIF;
	else if (keyword_match(op, "&message") == 0)		d = MESSAGE;
	else {
		warnx("%s:%d: invalid directive '%s'",
			scope->filename, scope->lineno, op);
		line->data[i] = save;
		return 0;
	}
	line->data[i] = save;
	if (i < line->len) i++;

	/* skip white space */
	while (i < line->len && isspace(line->data[i])) i++;

	if (d == GLOB || d == SCOP) {
		char *name, *text;
		int duplicate;
		/* struct buf etext; */

		name = line->data + i;
		while (i < line->len && !isspace(line->data[i])) i++;
		line->data[i] = '\0';
		if (i < line->len) i++;
		while (i < line->len && isspace(line->data[i])) i++;
		text = line->data + i;
		i = line->len - 1;
		while (i > text - line->data && isspace(line->data[i-1])) i--;
		line->data[i] = '\0';

		if (name == text || name == line->len + line->data) {
			warnx("%s:%d: missing arguments to &%s",
				scope->filename, scope->lineno, op);
			return 0;
		}

		if (!if_active())
			return 0;

		/* buf_init(&etext); */
		/* expand(scope, line, text - line->data, &etext, 0, 1); */
		/* text = etext.data; */
		/* etext.data[etext.len] = '\0'; */
#ifdef DEBUG
fprintf(stderr, "&%s '%s' <- '%s'\n", d == GLOB ? "GLOB" : "SCOP", name, text);
#endif
		if (d == GLOB)
			duplicate = scope_define(globalscope, name, text);
		else
			duplicate = scope_define(scope, name, text);
		if (duplicate)
			warnx("%s:%d: duplicate definition for %s",
				scope->filename, scope->lineno, name);
		/* buf_free(&etext); */
	}

	if (d == UNDEF && if_active()) {
		char *name;

		name = line->data + i;
		while (i < line->len && !isspace(line->data[i])) i++;
		line->data[i] = '\0';

		scope_undefine(scope, name);
	}

	if (d == MESSAGE && if_active()) {
		char *text = line->data + i;
		fprintf(stderr, "%s", text);
	}

	if (d == ELSE || d == ELSEIF) {
		if (scope->ifdepth == 0)
			warnx("%s:%d: &%s without &if",
				scope->filename, scope->lineno, op);
		else if (d == ELSE)
			if_else(1);
	}

	if (d == IF || d == ELSEIF) {
		int condition = 0;

		if ((d == IF && if_active()) ||
		    (d == ELSEIF && if_elseactive()))
		{
			condition = ifexpr(scope, line, i);
			if (condition == -1)	/* missing &THEN */
				return 1;
		}

		if (d == IF) {
			if_if(condition);
			scope->ifdepth++;
		} else
			if_else(condition);
	}

	if (d == ENDIF) {
		if (scope->ifdepth == 0) {
			warnx("%s:%d: too many &endifs",
				scope->filename, scope->lineno);
		} else {
			if_endif();
			scope->ifdepth--;
		}
	}

	return 0;
}

/* Main driver, treats command line as if it were {arg1 arg2 arg3 ...} */
int
main(argc, argv)
	int argc;
	char *argv[];
{
	int linesync = 1;
	int ch;
	extern int optind;

	if_init();

	while ((ch = getopt(argc, argv, "bP")) != -1)
		switch (ch) {
		case 'b':	batchmode = 1; break;
		case 'P':	linesync = 0; break;
		case '?':
		default:
			fprintf(stderr, "usage: %s [-bP] [file [args]]\n",
				argv[0]);
			exit(1);
		}
	argc -= optind;
	argv += optind;

	loadfile(NULL, argv, NULL, linesync);
	exit(0);
}
